/*
 * kgraft_patch_cve-2015-7613
 *
 * Fix for CVE-2015-7613, bsc#948701
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/ipc_namespace.h>
#include <linux/msg.h>
#include <linux/shm.h>
#include <linux/hugetlb.h>
#include <linux/mman.h>
#include <linux/shmem_fs.h>
#include <linux/file.h>

#include "ipc_util.h"

static void *(*kgr_ipc_rcu_alloc)(int size);
static void (*kgr_ipc_rcu_putref)(void *ptr, void (*func)(struct rcu_head *head));
static void (*kgr_ipc_rcu_free)(struct rcu_head *head);
static int (*kgr_security_msg_queue_alloc)(struct msg_queue *msq);
static void (*kgr_msg_rcu_free)(struct rcu_head *head);
static int (*kgr_security_shm_alloc)(struct shmid_kernel *shp);
static int *kgr_sysctl_overcommit_memory;
static void (*kgr_shm_rcu_free)(struct rcu_head *head);
static struct file *(*kgr_hugetlb_file_setup)(const char *name, size_t size, vm_flags_t acctflag,
			struct user_struct **user, int creat_flags, int page_size_log);
static const struct file_operations *kgr_hugetlbfs_file_operations;
static struct hstate (*kgr_hstates)[];
static unsigned int *kgr_default_hstate_idx;
static struct hstate *(*kgr_size_to_hstate)(unsigned long size);
static void (*kgr_user_shm_unlock)(size_t size, struct user_struct *user);
static int (*kgr_is_file_shm_hugepages)(struct file *file);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "ipc_rcu_alloc", (void *)&kgr_ipc_rcu_alloc },
	{ "ipc_rcu_putref", (void *)&kgr_ipc_rcu_putref },
	{ "ipc_rcu_free", (void *)&kgr_ipc_rcu_free },
	{ "security_msg_queue_alloc", (void *)&kgr_security_msg_queue_alloc },
	{ "msg_rcu_free", (void *)&kgr_msg_rcu_free },
	{ "security_shm_alloc", (void *)&kgr_security_shm_alloc },
	{ "sysctl_overcommit_memory", (void *)&kgr_sysctl_overcommit_memory },
	{ "shm_rcu_free", (void *)&kgr_shm_rcu_free },
	{ "hugetlb_file_setup", (void *)&kgr_hugetlb_file_setup },
	{ "hugetlbfs_file_operations", (void *)&kgr_hugetlbfs_file_operations },
	{ "hstates", (void *)&kgr_hstates },
	{ "default_hstate_idx", (void *)&kgr_default_hstate_idx },
	{ "size_to_hstate", (void *)&kgr_size_to_hstate },
	{ "user_shm_unlock", (void *)&kgr_user_shm_unlock },
	{ "is_file_shm_hugepages", (void *)&kgr_is_file_shm_hugepages },
};

#define IPC_SEM_IDS	0
#define IPC_MSG_IDS	1
#define IPC_SHM_IDS	2

#define sem_ids(ns)	((ns)->ids[IPC_SEM_IDS])
#define msg_ids(ns)	((ns)->ids[IPC_MSG_IDS])
#define shm_ids(ns)	((ns)->ids[IPC_SHM_IDS])

#define sc_semmsl	sem_ctls[0]
#define sc_semmns	sem_ctls[1]
#define sc_semopm	sem_ctls[2]
#define sc_semmni	sem_ctls[3]

#define SEQ_MULTIPLIER	(IPCMNI)
#define ipcid_to_idx(id) ((id) % SEQ_MULTIPLIER)
#define ipcid_to_seqx(id) ((id) / SEQ_MULTIPLIER)

#define kgr_default_hstate ((*kgr_hstates)[*kgr_default_hstate_idx])

static inline int kgr_ipc_buildid(int id, int seq)
{
	return SEQ_MULTIPLIER * seq + id;
}

static inline void kgr_ipc_unlock_object(struct kern_ipc_perm *perm)
{
	spin_unlock(&perm->lock);
}

int kgr_ipc_addid(struct ipc_ids* ids, struct kern_ipc_perm* new, int size)
{
	kuid_t euid;
	kgid_t egid;
	int id;
	int next_id = ids->next_id;

	if (size > IPCMNI)
		size = IPCMNI;

	if (ids->in_use >= size)
		return -ENOSPC;

	idr_preload(GFP_KERNEL);

	spin_lock_init(&new->lock);
	new->deleted = 0;
	rcu_read_lock();
	spin_lock(&new->lock);

	/*
	 * Fix CVE-2015-7613
	 *
	 * several lines moved here
	 */
	current_euid_egid(&euid, &egid);
	new->cuid = new->uid = euid;
	new->gid = new->cgid = egid;

	id = idr_alloc(&ids->ipcs_idr, new,
		       (next_id < 0) ? 0 : ipcid_to_idx(next_id), 0,
		       GFP_NOWAIT);
	idr_preload_end();
	if (id < 0) {
		spin_unlock(&new->lock);
		rcu_read_unlock();
		return id;
	}

	ids->in_use++;

	if (next_id < 0) {
		new->seq = ids->seq++;
		if (ids->seq > ids->seq_max)
			ids->seq = 0;
	} else {
		new->seq = ipcid_to_seqx(next_id);
		ids->next_id = -1;
	}

	new->id = kgr_ipc_buildid(id, new->seq);
	return id;
}

int kgr_newque(struct ipc_namespace *ns, struct ipc_params *params)
{
	struct msg_queue *msq;
	int id, retval;
	key_t key = params->key;
	int msgflg = params->flg;

	msq = kgr_ipc_rcu_alloc(sizeof(*msq));
	if (!msq)
		return -ENOMEM;

	msq->q_perm.mode = msgflg & S_IRWXUGO;
	msq->q_perm.key = key;

	msq->q_perm.security = NULL;
	retval = kgr_security_msg_queue_alloc(msq);
	if (retval) {
		kgr_ipc_rcu_putref(msq, kgr_ipc_rcu_free);
		return retval;
	}

	msq->q_stime = msq->q_rtime = 0;
	msq->q_ctime = get_seconds();
	msq->q_cbytes = msq->q_qnum = 0;
	msq->q_qbytes = ns->msg_ctlmnb;
	msq->q_lspid = msq->q_lrpid = 0;
	INIT_LIST_HEAD(&msq->q_messages);
	INIT_LIST_HEAD(&msq->q_receivers);
	INIT_LIST_HEAD(&msq->q_senders);

	/*
	 * Fix CVE-2015-7613
	 *
	 * several lines moved to the end of initialization
	 */
	/* ipc_addid() locks msq upon success. */
	id = kgr_ipc_addid(&msg_ids(ns), &msq->q_perm, ns->msg_ctlmni);
	if (id < 0) {
		kgr_ipc_rcu_putref(msq, kgr_msg_rcu_free);
		return id;
	}

	kgr_ipc_unlock_object(&msq->q_perm);
	rcu_read_unlock();

	return msq->q_perm.id;
}

static inline int kgr_is_file_hugepages(struct file *file)
{
	if (file->f_op == kgr_hugetlbfs_file_operations)
		return 1;
	if (kgr_is_file_shm_hugepages(file))
		return 1;

	return 0;
}

static inline struct hstate *kgr_hstate_sizelog(int page_size_log)
{
	if (!page_size_log)
		return &kgr_default_hstate;
	return kgr_size_to_hstate(1 << page_size_log);
}

int kgr_newseg(struct ipc_namespace *ns, struct ipc_params *params)
{
	key_t key = params->key;
	int shmflg = params->flg;
	size_t size = params->u.size;
	int error;
	struct shmid_kernel *shp;
	size_t numpages = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;
	struct file * file;
	char name[13];
	int id;
	vm_flags_t acctflag = 0;

	if (size < SHMMIN || size > ns->shm_ctlmax)
		return -EINVAL;

	if (ns->shm_tot + numpages > ns->shm_ctlall)
		return -ENOSPC;

	shp = kgr_ipc_rcu_alloc(sizeof(*shp));
	if (!shp)
		return -ENOMEM;

	shp->shm_perm.key = key;
	shp->shm_perm.mode = (shmflg & S_IRWXUGO);
	shp->mlock_user = NULL;

	shp->shm_perm.security = NULL;
	error = kgr_security_shm_alloc(shp);
	if (error) {
		kgr_ipc_rcu_putref(shp, kgr_ipc_rcu_free);
		return error;
	}

	sprintf (name, "SYSV%08x", key);
	if (shmflg & SHM_HUGETLB) {
		struct hstate *hs;
		size_t hugesize;

		hs = kgr_hstate_sizelog((shmflg >> SHM_HUGE_SHIFT) & SHM_HUGE_MASK);
		if (!hs) {
			error = -EINVAL;
			goto no_file;
		}
		hugesize = ALIGN(size, huge_page_size(hs));

		/* hugetlb_file_setup applies strict accounting */
		if (shmflg & SHM_NORESERVE)
			acctflag = VM_NORESERVE;
		file = kgr_hugetlb_file_setup(name, hugesize, acctflag,
				  &shp->mlock_user, HUGETLB_SHMFS_INODE,
				(shmflg >> SHM_HUGE_SHIFT) & SHM_HUGE_MASK);
	} else {
		/*
		 * Do not allow no accounting for OVERCOMMIT_NEVER, even
	 	 * if it's asked for.
		 */
		if  ((shmflg & SHM_NORESERVE) &&
				*kgr_sysctl_overcommit_memory != OVERCOMMIT_NEVER)
			acctflag = VM_NORESERVE;
		file = shmem_file_setup(name, size, acctflag);
	}
	error = PTR_ERR(file);
	if (IS_ERR(file))
		goto no_file;

	shp->shm_cprid = task_tgid_vnr(current);
	shp->shm_lprid = 0;
	shp->shm_atim = shp->shm_dtim = 0;
	shp->shm_ctim = get_seconds();
	shp->shm_segsz = size;
	shp->shm_nattch = 0;
	shp->shm_file = file;
	shp->shm_creator = current;

	/*
	 * Fix CVE-2015-7613
	 *
	 * several lines moved to the end of initialization
	 */
	id = kgr_ipc_addid(&shm_ids(ns), &shp->shm_perm, ns->shm_ctlmni);
	if (id < 0) {
		error = id;
		goto no_id;
	}

	/*
	 * shmid gets reported as "inode#" in /proc/pid/maps.
	 * proc-ps tools use this. Changing this will break them.
	 */
	file_inode(file)->i_ino = shp->shm_perm.id;

	ns->shm_tot += numpages;
	error = shp->shm_perm.id;

	kgr_ipc_unlock_object(&shp->shm_perm);
	rcu_read_unlock();
	return error;

no_id:
	if (kgr_is_file_hugepages(file) && shp->mlock_user)
		kgr_user_shm_unlock(size, shp->mlock_user);
	fput(file);
no_file:
	kgr_ipc_rcu_putref(shp, kgr_shm_rcu_free);
	return error;
}

static int kgr_patch_cve_2015_7613_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_cve_2015_7613_init(void)
{
	return kgr_patch_cve_2015_7613_kallsyms();
}
